import datetime

from pymongo import Connection
from pymongo.errors import ConnectionFailure
from pymongo.binary import Binary
from pymongo.errors import OperationFailure

from pyramid.exceptions import ConfigurationError
from pyramid.decorator import reify

from anykeystore.compat import pickle
from anykeystore.interfaces import KeyValueStore

def includeme(config):
    settings = config.registry.settings
    host = settings.get('velruse.store.host', 'localhost')
    port = int(settings.get('velruse.store.port', '27017'))
    db = settings.get('velruse.store.db')
    collection = settings.get('velruse.store.collection', 'velruse_ustore')

    if db is None:
        raise ConfigurationError('Missing "velruse.store.db" setting.')

    store = MongoDBStore(
        host=host, port=port, db=db, collection=collection,
    )
    config.registry.velruse_store = store

class MongoDBStore(KeyValueStore):
    """MongoDB Storage for Auth Provider"""
    def __init__(self, host='localhost', port=27017, db="mongo_db_name",
                 collection='velruse_ustore'):
        self.host = host
        self.port = port
        self.db = db
        self.collection = collection

    @reify
    def _conn(self):
        """The MongoDB connection, cached for this call"""
        try:
            db_conn = Connection(self.host, self.port, slave_okay=False)
        except ConnectionFailure:
            raise Exception('Unable to connect to MongoDB')
        conn = db_conn[self.db]
        #Set arbitrary limit on how large user_store session can grow to
        #http://www.mongodb.org/display/DOCS/Capped+Collections
        if not self.collection in conn.collection_names():
            conn.create_collection(self.collection,
                                   dict(capped=True, size=100000000))
        return conn

    def retrieve(self, key):
        data = self._conn[self.collection].find_one({'key': key})
        if data:
            return pickle.loads(data['value'])
        else:
            return None

    def store(self, key, value, expires=300):
        try:
            expire_time = datetime.datetime.utcnow() + \
                          datetime.timedelta(seconds=expires)
            self._conn[self.collection].update({'key': key},
                {'$set': {"value": Binary(pickle.dumps(value)),
                          "expires": expire_time}},
                upsert=True, safe=True)
        except OperationFailure:
            return False
        else:
            return True

    def delete(self, key):
        try:
            self._conn[self.collection].remove({'key': key})
        except OperationFailure:
            return False
        else:
            return True

    def purge_expired(self):
        pass



class KeyValueStore(object):
    """ Backend interface for storing and retrieving data."""

    def retrieve(self, key):
        """ This method retrieves the data for a key from the storage.

        :param key: The key to retrieve. Keys are always ascii-safe strings.
        """
        raise NotImplementedError

    def store(self, key, value, expires=None):
        """ This method stores value in the storage.

        For backend's that don't automatically expire data, some record should
        be kept with the key's data marking when it should have expired so that
        :meth:`~velruse.store.interface.purge_expired` can properly purge
        old data.

        :param key: The key to store the value under.
        :param value: The data to store.
        :param expires: Optional expiration time in seconds or a
                        :meth:`datetime.timedelta` object before the stored
                        data should be removed.
        """
        raise NotImplementedError

    def delete(self, key):
        """ This method deletes data from the storage.

        :param key: The key of the data to be removed.
        """
        raise NotImplementedError

    def purge_expired(self):
        """ This method purges all expired data from the storage.

        All expired data should be purged from the backend when this method
        is called. Backends that automatically expire old data must still
        implement this method, but can do nothing.
        """
        raise NotImplementedError



from datetime import datetime

from anykeystore.interfaces import KeyValueStore
from anykeystore.utils import coerce_timedelta

class MemoryStore(KeyValueStore):
    """ In-memory storage. Not persistent."""
    def __init__(self):
        self._store = {}

    def retrieve(self, key):
        data = self._store.get(key)
        if data:
            value, expires = data
            if expires is None or datetime.utcnow() < expires:
                return value

    def store(self, key, value, expires=None):
        expiration = None
        if expires:
            expiration = datetime.utcnow() + coerce_timedelta(expires)
        self._store[key] = (value, expiration)

    def delete(self, key):
        if key in self._store:
            del self._store[key]

    def purge_expired(self):
        now = datetime.utcnow()
        to_delete = []

        # Record the keys to delete, there may be a lot, so we use iteritems
        # which doesn't let us change it while iterating
        for key, value in self._store.iteritems():
            if value[1] is not None and now > value[1]:
                to_delete.append(key)
        for key in to_delete:
            del self._store[key]
        return True



import logging

try:
    import memcache
except ImportError:
    # fall back for Google App Engine -- hasnt been tested though
    from google.appengine.api import memcache

from anykeystore.compat import pickle
from anykeystore.interfaces import KeyValueStore


log = logging.getLogger(__name__)

#def includeme(config):
#    settings = config.registry.settings
#    servers = splitlines(settings.get('velruse.store.servers', ''))
#    key_prefix = settings.get('velruse.store.key_prefix', 'velruse_ustore')
#
#    if not servers:
#        raise ConfigurationError('Missing "velruse.store.servers" setting')
#
#    store = MemcachedStore(servers=servers, key_prefix=key_prefix)
#    config.registry.velruse_store = store

class MemcachedStore(KeyValueStore):
    """Memcached Storage for Auth Provider"""

    def __init__(self, servers=None, key_prefix='velruse_ustore'):
        self.key_prefix = key_prefix
        self.servers = servers or ['localhost:11211']

    @reify
    def _conn(self):
        """The Memcached connection, cached for this call"""
        return memcache.Client(self.servers)

    def _key(self, key):
        return str('%s:%s' % (self.key_prefix, key))

    def retrieve(self, key):
        return self._conn.get(self._key(key))

    def store(self, key, value, expires=None):
        log.debug('Servers %s storing %s=%s' % (
            self.servers, self._key(key), value))

        self._conn.set(self._key(key), value, expires or 0)
        return True

    def delete(self, key):
        log.debug('Deleting %s', key)
        self._conn.delete(self._key(key))

    def purge_expired(self):
        pass



import redis
from redis.exceptions import RedisError

from anykeystore.compat import pickle
from anykeystore.interfaces import KeyValueStore


def includeme(config):
    settings = config.registry.settings
    host = settings.get('velruse.store.host', 'localhost')
    port = int(settings.get('velruse.store.port', '6379'))
    db = int(settings.get('velruse.store.db', '0'))
    key_prefix = settings.get('velruse.store.key_prefix', 'velruse_ustore')
    store = RedisStore(
        host=host, port=port, db=db, key_prefix=key_prefix,
    )
    config.registry.velruse_store = store

class RedisStore(KeyValueStore):
    """Redis Storage for Auth Provider"""
    def __init__(self, host='localhost', port=6379, db=0,
                 key_prefix='velruse_ustore'):
        self.host = host
        self.port = port
        self.db = db
        self.key_prefix = key_prefix

    def _conn(self):
        """The Redis connection, cached for this call"""
        return redis.Redis(host=self.host, port=self.port, db=self.db)

    def retrieve(self, key):
        data = self._conn.get('%s:%s' % (self.key_prefix, key))
        if data:
            return pickle.loads(data)
        else:
            return None

    def store(self, key, value, expires=None):
        key = '%s:%s' % (self.key_prefix, key)
        try:
            self._conn.set(key, pickle.dumps(
                value, protocol=pickle.HIGHEST_PROTOCOL))
            if expires:
                self._conn.expire(key, expires)
        except RedisError:
            return False
        else:
            return True

    def delete(self, key):
        try:
            self._conn.delete('%s:%s' % (self.key_prefix, key))
        except RedisError:
            return False
        else:
            return True

    def purge_expired(self):
        pass



from datetime import datetime

from sqlalchemy import Column, MetaData
from sqlalchemy import String, Text, DateTime, Table
from sqlalchemy.sql import select, delete

from anykeystore.compat import pickle
from anykeystore.interfaces import KeyValueStore
from anykeystore.utils import coerce_timedelta


class SQLStore(KeyValueStore):
    """ Simple storage via SQLAlchemy.

    The store will automatically create a table object if one is not
    supplied. The automatically generated table is shown below. If a
    table is supplied directly, it must support the required columns
    `key`, `value`, and `expires`.

    .. code-block:: python
        table = Table(table_name, metadata,
            Column('key', String(200), primary_key=True, nullable=False),
            Column('value', Text(), nullable=False),
            Column('expires', DateTime()),
        )

    :param engine: A SQLAlchemy engine.
    :param table: Optional. The SQLAlchemy Table instance to be used for
                  storage. If this isn't supplied, a table is generated
                  automatically.
    :type table: sqlalchemy.Table
    :param table_name: Optional. The name of the table.
    :param metadata: Optional. The SQLAlchemy MetaData instance to hook the
                 generated table into.
    :type metadata: sqlalchemy.MetaData
    """
    def __init__(self, engine, **kw):
        self.engine = engine
        if 'table' in kw:
            self.table = kw['table']
        else:
            table_name = kw.pop('table_name', 'key_storage')
            meta = kw.pop('metadata', MetaData())
            self.table = Table(table_name, meta,
                Column('key', String(200), primary_key=True, nullable=False),
                Column('value', Text(), nullable=False),
                Column('expires', DateTime()),
            )

    def create(self):
        self.table.create(checkfirst=True, bind=self.engine)

    def retrieve(self, key):
        s = self.table
        data = self.engine.execute(
            select([s.c.value, s.c.expires], s.c.key == key))
        if data:
            value, expires = data
            if expires is None or datetime.utcnow() < expires:
                return pickle.loads(data[0])

    def store(self, key, value, expires=None):
        expiration = None
        if expires:
            expiration = datetime.utcnow() + coerce_timedelta(expires)
        value = pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL)
        self.engine.execute(
            self.table.insert(),
            key=key, value=value, expires=expiration)

    def delete(self, key):
        self.engine.execute(delete(self.table, self.table.c.key == key))

    def purge_expired(self):
        self.engine.execute(
            delete(self.table, self.table.c.expires < datetime.utcnow()))



from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()
except IOError:
    README = CHANGES = ''

requires = []

setup(
    name='anykeystore',
    version='0.1a1',
    description="A key-value store supporting multiple backends.",
    long_description=README + '\n\n' + CHANGES,
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Topic :: Database',
    ],
    keywords='',
    author='Michael Merickel',
    author_email='oss@m.merickel.org',
    url='',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    entry_points="""\
    """,
)
